<?php

namespace Nwb\UrlToPdf;

class UrlToPdf
{
    /**
     * @var string
     */
    private string $apiUrl = 'https://url-to-pdf.kub.nwb.fr';

    /**
     * @var string|null
     */
    private ?string $accessKey;

    /**
     * UrlToPdf constructor.
     * @param array $options Liste des options de configuration
     * @throws Exception
     */
    public function __construct(array $options = [])
    {
        $key = $_ENV['URL_TO_PDF_ACCESS_KEY'] ?? $options['accessKey'] ?? null;

        if (!$key) {
            throw new \Exception('Missing access key');
        }

        $this->accessKey = $key;

        if (!empty($options['apiUrl'])) {
            $this->apiUrl = $options['apiUrl'];
        }
    }

    /**
     * Requête curl vers l'API
     * 
     * @param string $url vers la page html à convertir en PDF
     * @param array $options Liste des options de configuration
     * @return false|resource contenu du PDF
     * @throws Exception si la page est vide
     * @throws Exception si le code HTTP n'est pas 200
     * @throws Exception si le contenu de la réponse n'est pas un PDF
     * @throws Exception si le contenu de la réponse est vide
     */
    private function request(string $url, array $options)
    {
        // Fusion des options par défaut avec les options fournies

        $ch = curl_init();

        $pageContent = file_get_contents($options['url']);

        // Vérifier si la page est vide
        if (empty($pageContent)) {
            throw new \Exception("La page est vide.");
        }

        $headers = [
            'Content-Type: application/json',
            'access-key: ' . $this->accessKey,
        ];

        curl_setopt($ch, CURLOPT_URL, $this->apiUrl . $url);
        curl_setopt($ch, CURLOPT_POST, true);
        curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode($options));
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);

        $response = curl_exec($ch);

        $httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        $contentType = curl_getinfo($ch, CURLINFO_CONTENT_TYPE);

        if ($httpCode !== 200) {
            if (strpos($contentType, 'application/json') !== false) {
                $errorText = json_decode($response, true);

                if (count($errorText)) {
                    throw new \Exception($errorText);
                }
            } else {
                if ($response) {
                    throw new \Exception($response);
                }
            }

            throw new \Exception('Error: ' . $httpCode);
        }

        if (!$response || substr($response, 0, 4) !== "%PDF") {
            throw new \Exception('No body');
        }

        return $response;
    }

    /**
     * Enregistre le contenu dans un fichier sur le disque.
     *
     * @param string $content Contenu à enregistrer.
     * @param string $destination Chemin de destination du fichier.
     * @return bool
     */
    public function saveTo(string $content, string $destination): bool
    {
        // Écriture du contenu dans le fichier
        return file_put_contents($destination, $content) !== false;
    }

    /**
     * Retourne un PDF à partir d'une URL et des options
     * 
     * @param array $options Liste des options de configuration
     * @return false|resource contenu du PDF
     */
    public function getPdf(?array $options)
    {
        $options = $options ?? [
            'url' => 'https://normandiewebschool.fr/cursus/bachelor-developpement-web/',
            'loadingOptions' => ['waitUntil' => 'networkidle2'],
            'pdfOptions' => ['format' => 'a4']
        ];

        return $this->request('/get-pdf', $options);
    }

    /**
     * Retourne un PDF à partir d'un contenu HTML
     * 
     * @param string $htmlContent type TCPDF qui peut contenir des liens vers des images/fichiers sur le disque (ex: src:"/var/www/html/public-data/tmp/img_to_pdf/mon_image.jpg")
     * @param string $projectRootPath Chemin racine du projet (ex: /var/www/html/)
     * @param string $webPath Chemin vers le répertoire web (ex: /var/www/html/public-data/tmp/img_to_pdf/)
     * @param string $webPathUrl Chemin vers le répertoire web accessible depuis l'API (ex: https://www.nwb.fr/public-data/tmp/img_to_pdf/)
     * @param bool $isTestMode Indique si on est en mode test, dans ce cas on utilise l'API pour envoyer le fichier dans l'url en Base64
     * @return false|resource contenu du PDF
     * @throws Exception
     * 
     */
    public function getPdfFromHtml(string $htmlContent, string $projectRootPath, string $webPath, string $webPathUrl, bool $isTestMode = false): string
    {
        // Si répertoire web n'existe pas on trhow une exception
        if (!is_dir($webPath)) {
            throw new \Exception("Le répertoire web n'existe pas : " . $webPath);
        }

        // Test si $webPathUrl existe et fait plus de 8 caractères, si non on throw une exception
        if (!$webPathUrl || strlen($webPathUrl) < 8) {
            throw new \Exception("Le chemin web n'est pas valide : " . $webPathUrl);
        }

        // Test si $webPath se termine par un slash, si non on le rajoute
        if (substr($webPath, -1) !== '/') {
            $webPath .= '/';
        }

        // Test si $webPathUrl se termine par un slash, si non on le rajoute
        if (substr($webPathUrl, -1) !== '/') {
            $webPathUrl .= '/';
        }

        // Test si $webPathUrl ne commence pas par https:// ou http:// , si non on le rajoute
        if (strpos($webPathUrl, 'https://', 0) !== 0 && strpos($webPathUrl, 'http://', 0) !== 0) {
            $webPathUrl = 'https://' . $webPathUrl;
        }

        // Adapte le html pour accessibilité web 
        $htmlContent = $this->getHtmlWithCleanPath($htmlContent, $projectRootPath, $webPath, $webPathUrl);

        // On hash le contenu html pour créer un nom de fichier unique
        $fileName = md5($htmlContent) . '.pdf';

        // On créer un sous dossier pour stocker les PDF
        $webPath .= 'pdf/';

        // On créer le répertoire s'il n'existe pas
        if (!is_dir($webPath)) {
            mkdir($webPath, 0777, true);
        }

        // On créer le fichier
        $filePath = $webPath . $fileName;
        $file = fopen($filePath, 'w');

        // Enregistre le html dans un fichier temporaire
        fwrite($file, $htmlContent);
        fclose($file);

        // Appel de l'API getPdf
        if (!$isTestMode) {
            $pdf = $this->getPdf([
                'url' => $webPathUrl . 'pdf/' . $fileName,
                'pdfOptions' => ['format' => 'a4']
            ]);
        }
        // Si on est en mode test on passe une url de test
        if ($isTestMode) {
            $htmlContent = base64_encode($htmlContent);
            // Troncature pour que le html content fasse moins de 3000 caractères
            $htmlContent = substr($htmlContent, 0, 3000);
            $pdf = $this->getPdf([
                'url' => 'https://n8n.kub.nwb.fr/webhook/Base64ToHtml?htmlB64=' . $htmlContent,
                'pdfOptions' => ['format' => 'a4']
            ]);
        }

        // Remplacement du contenu du fichier temporaire par le contenu du PDF
        $file = fopen($filePath, 'w');
        fwrite($file, $pdf);
        fclose($file);

        // Retourne le path temporaire du PDF
        return $filePath;
    }


    /**
     * Retourne le html passé en paramètre avec les path des fichiers accessibles pour l'API Url to PDF
     * 
     * @param string $htmlIn
     * @param string $projectRootPath (ex: /var/www/html/)
     * @param string $webPath (ex: /var/www/html/public-data/tmp/img_to_pdf/)
     * @param string $webPathUrl (ex: https://www.nwb.fr/public-data/tmp/img_to_pdf/)
     * @return string $htmlContent
     * @throws Exception si l'image n'existe pas
     */
    public function getHtmlWithCleanPath(string $htmlIn, string $projectRootPath, string $webPath, string $webPathUrl): string
    {
        if (!$htmlIn) {
            throw new \Exception("Le contenu HTML est vide.");
        }
        if (!$projectRootPath) {
            throw new \Exception("Le chemin racine du projet n'est pas renseigné.");
        }
        if (!$webPath) {
            throw new \Exception("Le chemin vers le répertoire web n'est pas renseigné.");
        }
        if (!$webPathUrl) {
            throw new \Exception("L'url vers le répertoire web n'est pas renseigné.");
        }

        // Test si $webPath se termine par un slash, si non on le rajoute
        if (substr($webPath, -1) !== '/') {
            $webPath .= '/';
        }

        // Test si $webPathUrl se termine par un slash, si non on le rajoute
        if (substr($webPathUrl, -1) !== '/') {
            $webPathUrl .= '/';
        }

        // On récupère le contenu HTML
        $htmlContent = $htmlIn;

        // On parse le contenu HTML
        $dom = new \DOMDocument();
        $dom->loadHTML($htmlContent);

        // On définit les tags et attributs à traiter
        $tagAttributes = [
            'img' => 'src',
            'script' => 'src',
            'link' => 'href',
        ];

        // On parcourt tous les éléments à traiter
        foreach ($tagAttributes as $tag => $attribute) {
            $elements = $dom->getElementsByTagName($tag);
            if (empty($elements)) {
                continue;
            }

            foreach ($elements as $element) {
                $filePath = $element->getAttribute($attribute);

                // Test si le chemin est vide
                if (!$filePath) {
                    continue;
                }

                // Si le chemin contient ./ ou ../ on les supprime
                $filePath = str_replace('../', '', $filePath);
                $filePath = str_replace('./', '', $filePath);

                // Si le chemin est une url ou un protocole de fichier ftp ou s3 on supprime la partie protocole et domaine
                $formatPath = "/^(https?|ftp|s3):\/\/([\da-z\.-]+)\.([a-z\.]{2,6})([\/\w \.-]*)*\/?$/";
                if (preg_match($formatPath, $filePath)) {
                    $filePath = parse_url($filePath, PHP_URL_PATH);
                }

                // Si le chemin commence par un slash et contient le pathRoot on le supprime
                if (strpos($filePath, '/', 0) === 0 && strpos($filePath, $projectRootPath, 0) !== 0) {
                    $filePath = substr($filePath, 1);
                }

                // Test du chemin réel
                if (realpath($filePath)) {
                    $filePath = realpath($filePath);
                }

                // On ajoute le chemin racine du projet s'il n'est pas déjà présent
                if (strpos($filePath, $projectRootPath, 0) !== 0) {
                    $filePath = $projectRootPath . '/' . $filePath;
                }

                // On test si le fichier existe
                if (!file_exists($filePath)) {
                    throw new \Exception("Le fichier n'existe pas : " . $filePath);
                }

                // Nouveau chemin absolu
                $newFilePath = $webPath . basename($filePath);

                // On créer le répertoire s'il n'existe pas
                if (!is_dir(dirname($newFilePath))) {
                    mkdir(dirname($newFilePath), 0777, true);
                }

                // On copie le fichier dans le nouveau répertoire
                if (!copy($filePath, $newFilePath)) {
                    throw new \Exception("Le fichier n'a pas pu être copié.");
                }

                // On remplace le chemin dans l'attribut de l'élément par le chemin accessible depuis l'API
                $formatPath = "/^(https?|ftp|s3):\/\/([\da-z\.-]+)\.([a-z\.]{2,6})([\/\w \.-]*)*\/?$/";
                $newRelativePath = $webPathUrl . basename($filePath);
                if (!preg_match($formatPath, $newRelativePath)) {
                    throw new \Exception("L'url n'est pas valide pour : " . $newRelativePath);
                }
                $element->setAttribute($attribute, $newRelativePath);
            }
        }

        // On récupère le nouveau contenu HTML
        $newHtmlContent = $dom->saveHTML();

        return $newHtmlContent;
    }
}
